
from __future__ import print_function
import sys
import os


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def normalize_dependencies(deps):
    if isinstance(deps, dict):
        for dep_name, dep in deps.items():
            for key in ["url", "sha256"]:
                if key not in dep:
                    raise Exception("error: dependency {} missing {}".format(dep_name, key))
    elif isinstance(deps, list):
        deps_list = deps
        deps = {}
        for dep in deps_list:
            for key in ["name", "url", "sha256"]:
                if key not in dep:
                    raise Exception("error: dependency missing {}".format(key))
            deps[dep["name"]] = { k:v for k, v in dep.items() if k != "name" }
    else:
        raise Exception("error: invalid dependencies format")
    return deps




verbose = False
debug = False

def set(args=None):
    global verbose, debug
    if args is None:
        return
    if 'debug' in args:
        debug = args['debug']
    if 'verbose' in args:
        verbose = args['verbose'] or debug

def dump():
    print("verbose: ", verbose)
    print("debug: ", debug)



import json
import re
from zipfile import ZipFile, BadZipfile
from typing import Dict, Any, IO, Tuple, Optional  # noqa: F401
from .common import *


INVALID_METADATA = "module metadata invalid"
DEFAULT_MAX_MODULE_FILE_SIZE = 1024 * 1024 * 10


class UnpackerPackageError(Exception):
    """
    Represents an error within the unpacking process
    """

    def __init__(self, message, reason=None, error_code=None, error_details=None):
        # type: (str, Optional[str], Optional[str], Dict[str, Any]) -> None
        super(UnpackerPackageError, self).__init__(message)
        self.reason = reason
        self.error_code = error_code
        self.error_details = error_details

    def __str__(self):
        # type: () -> str
        return "{}, reason: {}".format(super(UnpackerPackageError, self).__str__(), self.reason)


def unpack(bundle, max_bundle_size_mb=None):
    # type: (IO[bytes], Optional[int]) -> Tuple[Dict[str, Any], IO[bytes]]
    """
    Unpacks a bundled module, performs sanity validation on bundle.
    both the module metadata and the actual module are returned
    :rtype: tuple
    """
    try:
        with ZipFile(bundle) as zf:
            _validate_zip_file(zf, max_bundle_size_mb)
            metadata = json.load(zf.open('module.json'))
            module = zf.open(metadata["module_file"])
            _validate_metadata(metadata)

    except BadZipfile:
        raise UnpackerPackageError(message="Failed to extract bundle")

    except (IOError, ValueError):
        raise UnpackerPackageError("Failed to read module.json")

    except UnpackerPackageError:  # re-raise exceptions raised by validator-methods
        raise

    return metadata, module


def _validate_zip_file(zip_file, max_bundle_size_mb):
    # type: (ZipFile, Optional[int]) -> None
    """
    Checks if all entries within the zip file don't
    exceed a certain threshold.
    """
    infolist = zip_file.infolist()
    # We're expecting exactly two files.
    if len(infolist) != 2:
        raise UnpackerPackageError(message="module zip file content invalid",
                                   reason="module zip file should contains exactly two file",
                                   error_code="invalid_number_of_files")

    file_size_limit_bytes = max_bundle_size_mb * 1024 * 1024 if max_bundle_size_mb else DEFAULT_MAX_MODULE_FILE_SIZE

    # Check zip content size.
    for zip_info in infolist:
        # Size of the compressed/uncompressed data.
        if zip_info.file_size > file_size_limit_bytes:
            raise UnpackerPackageError(message="module zip file did not pass sanity validation",
                                       reason="module file content is too big",
                                       error_code="module_size_too_big",
                                       error_details={'module_size': zip_info.compress_size,
                                                      'max_size': file_size_limit_bytes})


def _validate_metadata(metadata):
    # type: (Dict[str, Any]) -> None
    """
    Checks metadata isn't missing any required fields
    :param metadata: dictionary
    :raises: UnpackerPackageError
    """

    if not metadata["module_name"]:
        raise UnpackerPackageError(message=INVALID_METADATA,
                                   reason="Empty module name",
                                   error_code="module_name_missing")

    if not metadata["module_file"]:
        raise UnpackerPackageError(message=INVALID_METADATA,
                                   reason="Empty module file name",
                                   error_code="module_file_missing")

    if metadata["architecture"] != 'x86_64':
        raise UnpackerPackageError(message=INVALID_METADATA,
                                   reason="Architecture must be 64 bits",
                                   error_code="module_architecture_not_supported",
                                   error_details={'expected': 'x86_64'})

    if not metadata["version"]:
        raise UnpackerPackageError(message=INVALID_METADATA,
                                   reason="Missing version",
                                   error_code="module_version_missing")

    try:
        metadata["version"] = int(float(metadata["version"]))
    except (ValueError, TypeError):
        raise UnpackerPackageError(message=INVALID_METADATA,
                                   reason="Module version should be an integer",
                                   error_code="module_version_not_integer")

    if "os_list" in metadata:
        try:
            metadata["os_list"] = [_os_version_parser(os) for os in metadata["os_list"]]
        except UnpackerPackageError:
            raise

    # wrong signature
    # TODO: this check should be deferred to a later stage
    # As sha256_checksum will read entire module file
    # And we're unable to seek it back to its starting point.
    # if sha256_checksum(module) != metadata["sha256"]:
    # [If needed, use module_metadata.sha256_checksum]
    #     raise UnpackerPackageError(message="module did not pass sanity validation",
    #                                reason="Wrong signature")


def _os_version_parser(os_version):
    # type: (str) -> Dict[str, str]
    """
    Parses an OS version string from the RAMP file to the same format as in the distro package
    :param os_version: a string in the format of the RAMP file, e.g. 'ubuntu18.04'
    :return: dictionary with the keys 'name' and 'version', for the OS name and OS version.
    """
    try:
        name_matcher = re.search(r'[a-zA-Z_]+', os_version)
        version_matcher = re.search(r'(\d+)(\.\d+)?(\.\d+)?', os_version)
        assert name_matcher and version_matcher
        name = name_matcher.group(0)
        version = version_matcher.group(0)
    except (ValueError, TypeError, AssertionError):
        raise UnpackerPackageError(message=INVALID_METADATA,
                                   reason="Can't parse OS '{}'".format(os_version),
                                   error_code="module_os_list_invalid")
    return {'name': name, 'version': version}



import os
import json
import zipfile
import yaml
import semantic_version
import tempfile
from subprocess import Popen, PIPE

from RAMP import config
import RAMP.module_metadata as module_metadata
from RAMP.commands_discovery import discover_modules_commands
from .common import *


def version_to_semantic_version(version):
    """
    Creates a semantic version from a numeric redis module version.
    """

    major = int(version / 10000)
    version -= major * 10000
    minor = int(version / 100)
    version -= minor * 100
    patch = int(version)

    sem_version_str = '{}.{}.{}'.format(major, minor, patch)
    # sem_version_str = '%02d.%02d.%02d' % (major, minor, patch)
    return semantic_version.Version(sem_version_str)

def set_defaults(module_path):
    """
    Creates a module metadata using default values
    """
    metadata = module_metadata.create_default_metadata(module_path)
    return metadata

def init_from_manifest(metadata, manifest):
    """
    Creates module metadata from user provided menifest file
    """
    try:
        data = yaml.load(manifest)
        for key, val in data.items():
            if key in metadata:
                metadata[key] = val
            else:
                eprint('{} unknown attribute'.format(key))
    except yaml.YAMLError as exc:
        eprint(exc)


def archive(module_path, metadata, archive_name='module.zip'):
    """
    Archives both module and module metadata.
    """
    with tempfile.NamedTemporaryFile(mode='w', prefix='ramp.json', delete=False) as outfile:
        jfile = outfile.name
        archive_name = archive_name.format(**metadata)
        json.dump(metadata, outfile, indent=4, sort_keys=True)

    try:
        with zipfile.ZipFile(archive_name, 'w', zipfile.ZIP_DEFLATED) as archive_file:
            archive_file.write(module_path, metadata["module_file"])
            archive_file.write(jfile, arcname='module.json')
            print(archive_name)
    finally:
        os.remove(jfile)


def package(module, **args):
    module_path = module

    nonkeys = dict.fromkeys(['manifest', 'verbose', 'print_filename_only', 'packname_file', 'output'], 1)
    manifest = args['manifest']
    print_filename_only = args['print_filename_only']
    packname_file = args['packname_file']
    output = args['output']

    # start with default values (lowest priority)
    metadata = set_defaults(module_path)

    # fill in keys from manifest file
    if manifest:
        init_from_manifest(metadata, manifest)

    # fill in keys from arguments
    for key in args.keys():
        if key in nonkeys:
            continue
        value = args[key]
        if value is None or value == []:
            continue
        metadata[key] = value

    # Load module into redis and discover its commands
    cmd_line_args = metadata.pop('run_command_line_args', None)
    module = discover_modules_commands(module_path, cmd_line_args if cmd_line_args else metadata["command_line_args"])
    metadata["module_name"] = module.name
    metadata["version"] = module.version
    metadata["semantic_version"] = str(version_to_semantic_version(module.version))
    metadata["commands"] = [cmd.to_dict() for cmd in module.commands if cmd.command_name not in metadata["exclude_commands"]]

    # fill in keys from arguments once more (highest prioriry)
    for key in args.keys():
        if key in nonkeys:
            continue
        value = args[key]
        if value is None or value == []:
            continue
        metadata[key] = value

    # override requested commands data
    for override in metadata["overide_command"]:
        if 'command_name' not in override:
            eprint("error: the given override json does not contains command name: %s" % str(override))
            continue
        override_index = [i for i in range(len(metadata["commands"])) if metadata["commands"][i]['command_name'] == override['command_name']]
        if len(override_index) != 1:
            eprint("error: the given override command appears more then once")
            continue
        if config.verbose:
            print('overiding %s with %s' % (str(metadata["commands"][override_index[0]]), str(override)))
        metadata["commands"][override_index[0]] = override

    for add in metadata["add_command"]:
        if 'command_name' not in add:
            eprint("error: the given add command json does not contains command name: %s" % str(add))
            continue
        if config.verbose:
            print('adding command %s' % (str(add)))
        metadata["commands"].append(add)


    module_name = args['module_name']
    if module_name:
        metadata["module_name"] = module_name

    # validate and normalize dependencies
    metadata["dependencies"] = normalize_dependencies(metadata["dependencies"])
    metadata["optional-dependencies"] = normalize_dependencies(metadata["optional-dependencies"])

    try:
        p = Popen('git rev-parse HEAD'.split(' '), stdin=PIPE, stdout=PIPE, stderr=PIPE)
        git_sha, err = p.communicate()
        if p.returncode != 0:
            eprint("could not extract git sha {}".format(err))
        else:
            metadata["git_sha"] = str(git_sha.strip())
    except Exception:
        eprint("could not extract git sha {}".format(err))

    # cleanup metadata from utility keys
    fields = dict.fromkeys(module_metadata.FIELDS, 1)
    for key in list(metadata.keys()):
        if not key in fields:
            metadata.pop(key)

    packname = output.format(**metadata)
    if packname_file:
        with open(packname_file, 'w') as file:
            file.write(packname)

    if print_filename_only:
        # For scripts, it might be helpful to know the formatted filename
        # ahead of time, so that it can manipulate it later on.
        if not packname_file:
            print(packname)
        return 0

    if config.verbose:
        print("Module Metadata:")
        print(json.dumps(metadata, indent=2))

    archive(module_path, metadata, archive_name=output)
    return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['RAMP', 'RAMP.disposableredis']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'click>=6.7.0,<7.0.0',
 'redis>=3.5.3,<4.0.0',
 'semantic-version>=2.8.5,<3.0.0',
 'typing>=3.0.0,<4.0.0']

entry_points = \
{'console_scripts': ['ramp = RAMP.ramp:ramp']}

setup_kwargs = {
    'name': 'ramp-packer',
    'version': '2.1.0',
    'description': 'Packs for Redis modules into a distributable format',
    'long_description': "[![license](https://img.shields.io/github/license/RedisLabsModules/RAMP.svg)](https://github.com/RedisLabsModules/RAMP)\n[![CircleCI](https://circleci.com/gh/RedisLabsModules/RAMP/tree/master.svg?style=svg)](https://circleci.com/gh/RedisLabsModules/RAMP/tree/master)\n[![GitHub issues](https://img.shields.io/github/release/RedisLabsModules/RAMP.svg)](https://github.com/RedisLabsModules/RAMP/releases/latest)\n[![Codecov](https://codecov.io/gh/RedisLabsModules/RAMP/branch/master/graph/badge.svg)](https://codecov.io/gh/RedisLabsModules/RAMP)\n\n\n# RAMP\n\nRedis Automatic Module Packaging\n\nSimilar to `npm init`, this packer bundles Redis Modules for later distribution.\n\nIt gathers information from modules e.g.\nmodule's name, command list, version and additional metadata.\n\n## Prerequisites\n\nMake sure redis-server is on your PATH\n\n```sh\nexport PATH=$PATH:<PATH_TO_REDIS>\n```\n\n## Install\n\nYou can either use pip install or the setup.py script\n\n```sh\npip install ramp-packer\n```\n\n```sh\npython setup.py install\n```\n\n## Usage\n\n## Manifest mode\n\n```sh\nramp pack <PATH_TO_RedisModule.so> -m <PATH_TO_Manifest.yml>\n```\n\nmanifest.yml should specify your module's attributes, the ones you would specify manualy if you were to use\nthe Command line mode, see Full usage options and manifest.yml for a reference.\n\n## Command line mode\n\n```sh\nramp pack <PATH_TO_RedisModule.so> -a <author> -e <email> -A <architecture> -d <description> -h <homepage> -l <license> -c <cmdargs> -r <redis-min-version>\n```\n\n## Full usage options\n\n```sh\nUsage: ramp [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  pack\n  unpack\n  validate\n  version\n```\n\n## Packing\n\n```sh\nUsage: ramp pack [OPTIONS] MODULE\n\nOptions:\n  -o, --output TEXT               output file name\n  -v, --verbose                   verbose mode: print the resulting metadata\n  -m, --manifest FILENAME         generate package from manifest\n  -d, --display-name TEXT         name for display purposes\n  -a, --author TEXT               module author\n  -e, --email TEXT                authors email\n  -A, --architecture TEXT         module compiled on i386/x86_64 arch\n  -D, --description TEXT          short description\n  -h, --homepage TEXT             module homepage\n  -l, --license TEXT              license\n  -c, --cmdargs TEXT              module command line arguments\n  -r, --redis-min-version TEXT    redis minimum version\n  -R, --redis-pack-min-version TEXT\n                                  redis pack minimum version\n  -cc, --config-command TEXT      command to configure module at run time\n\n  -O, --os TEXT                   build target OS (Darwin/Linux)\n  -C, --capabilities TEXT         comma seperated list of module capabilities\n  --help                          Show this message and exit.\n```\n\n## Module Capabilities\n\nFollowing is a list of capabilities which can be specified for a module\n\ncapability | description |\n---------- | ----------- |\ntypes | module has its own types and not only operate on existing redis types|\nno_multi_key | module has no methods that operate on multiple keys|\nreplica_of | module can work with replicaof capability when it is loaded into a source or a destination database|\nbackup_restore | module can work with import/export capability|\neviction_expiry | module is able to handle eviction and expiry without an issue|\nreshard_rebalance | module is able to operate in a database that is resharded and rebalanced|\nfailover_migrate | module is able to operate in a database that is failing over and migrating|\npersistence_aof | module is able to operate in a database when database chooses AOF persistence option|\npersistence_rdb | module is able to operate in a database when database chooses SNAPSHOT persistence option|\nhash_policy | module is able to operate in a database with a user defined HASH POLICY|\nflash | module is able to operate in a database with Flash memory is enabled or changed over time|\ncrdb | module is able to operate in a database with crdt for the default redis data types|    \nclustering | module is able to operate in a database that is sharded and shards can be migrated|\nintershard_tls | module supports two-way encrypted communication between shards|\n\n## Output\n\nramp pack generates module.zip\n\nWhich contains:\n\n    1. RedisModule.so - original module\n    2. Module.json - module's metadata\n\n## Test\nMake sure redis-server is on your PATH\n\n```sh\nexport PATH=$PATH:<PATH_TO_REDIS>\n```\n\nInstall RAMP\n```sh\npython setup.py install\n```\n\nCompile RedisGraph for your OS v1.0.12 (https://github.com/RedisLabsModules/RedisGraph/tree/v1.0.12)\n\nCopy `redisgraph.so` in `test_module` directory in the root of this project.\n\nRun tests\n```sh\npython test.py\n```\n",
    'author': 'RedisLabs',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)
